/*
 * coscr.c
 *
 * Copyright (c) 1997-1999,2000,2001 Hajime Mochizuki <motizuki@jaist.ac.jp>
 *   Natural Language Processing Laboratory,
 *   Japan Advanced Institute of Science and Technology.
 *
 *
 * Use of this software is granted only for RESEARCH PURPOSES. 
 * Redistribution is granted as long as all copyright notices are retained.
 * Redistribution with modification is allowed provided that you make your 
 * modified version obviously distinguishable from the original one.  
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR 
 * IMPLIED WARRANTIES ARE DISCLAIMED.
 *
 * geta б
 */
#include <config.h>


#ifdef BSD44
#include <sys/param.h>
#endif
#include <stdio.h>
#include <string.h>
#include <math.h>

#include <assert.h>
#include <geta/paths.h>
#include <geta/config.h>
#include <geta/wam.h>

#include "coscr.h"
#include "mrpfil.h"

double threshold;
struct _copair  copair;

WAM *s_wam_init_h (handle)
    char *handle;
{
	char getaroot[200];
	getaroot[0]='\0';
	
	if (!(getenv(PROOTENV))) {
		wam_init (DEF_GETAROOT);
	} else {
		wam_init (NULL);
	}
	return wam_open(handle,0);
}

void s_wam_finish (w)
    WAM *w;
{
	wam_close(w);
}    

double calc_co_score (df1, df2, df, calc_mode, N)
    int df1;
    int df2;
    int df;
    int calc_mode;
    int N;
{
	double bb, bs, scr, tmp;

	switch (calc_mode) {
	    case 1: /* tf idf like ʤ*/

		bs = (double)df;
		if (df2==0 || df1==0) {
			fprintf (stderr, "Warning: df1 or df2 == 0\n");
			scr = (double)0;
		}else {
			bb = log(((double)N*(double)N)/((double)df2*(double)df1));
			scr = bs * bb;
		}

		if (threshold == (double)0) threshold = (double)500;
		break;

	    case 2:
		/* mutual information 
		 * scr (t1, t2) = {df (t1, t2) * N} / {df(t1) * df(t2)} */

		bs = (double)df * (double)N;
		if (df1==0 || df2==0) {
			fprintf (stderr, "Warning: df1 or df2 == 0\n");
			scr = (double)0;
		} else {
			bb = (double)df1 * (double)df2;
			scr = bs / bb;
		}
		if (threshold == (double)0) threshold = (double)500;
		break;

	    case 3:
		/* tf idf like t1ߤt2Ψ
		 * scr (t1, t2) = df(t1,t2)/ log ( N/df(t2) ) */

		bs = (double)df;
		if (df2 == 0) {
			fprintf (stderr, "Warning: df2=0\n");
			scr = (double)0;
		} else {
			bb = log((double)N/(double)df2);
			scr = bs * bb;
		}
		if (threshold == (double)0) threshold = (double)500;
		break;

	    case 4:
		/* tf idf like t2ߤt1Ψ
		 * scr (t1, t2) = df(t1,t2)/ log ( N/df(t1) ) */
		if (df1==0) {
			fprintf (stderr, "Warning: df1=0\n");
			scr = (double)0;
		} else {
			bb = log((double)N/(double)df1);
			bs = (double)df;
			scr = bs * bb;
		}
		if (threshold == (double)0) threshold = (double)500;
		break;

	    case 5:
		/* mutual information t1ߤt2Ψ */
		bs = (double)df * (double)N;
		if (df2==0) {
			fprintf (stderr, "Warning: df2=0\n");
			scr = (double)0;
		} else {
			bb = (double)df2;
			scr = bs / bb;
		}
		if (threshold == (double)0) threshold = (double)500;
		break;
		
	    case 6:
		/* mutual information t2ߤt1Ψ */
		bs = (double)df * (double)N;
		bb = (double)df1;
		if  (df1==0) {
			fprintf (stderr, "Warning: df1=0\n");
			scr = (double)0;
		} else {
			scr = bs / bb;
		}
		if (threshold == (double)0) threshold = (double)500;
		break;

	    default:
		bs = (double)df;
		if (df2==0 || df1==0) {
			fprintf (stderr, "Warning: df1 or df2 =0\n");
			scr = (double)0;
		} else {
			bb = log (((double)N * (double)N)/((double)df2 * (double)df1));
			scr = bs * bb;
		}
		if (threshold == (double)0) threshold = (double)500;
		break;
	}
	return (scr);
}
    

int calc_coscr (handle, type, t, calc_mode, N, w)
    char *handle;
    int   type;
    int  *t;
    int   calc_mode; /* ׻Υ⡼ */
    int   N;
    WAM   *w;
{
	double scr;
    
	symbol_t id;
	ival_t elem_num;
	xr_vec *vec;

	int nrecords[2];  /* type = WAM_W| XR_W λtermdf+1ˤʤ */
	int freq_sum;
	int nidxsize;
	int r1, r2, r;
	int n, i;
	int zeroflag;

	int freq;
	symbol_t **doc;
	symbol_t **tf;
	
	double xi, yi, xy;

	if ((doc=(symbol_t **)calloc(2, (size_t)sizeof(symbol_t*)))==NULL) {
		fprintf (stderr, "malloc failed\n");
		exit(1);
	}
	if ((tf =(symbol_t **)calloc(2, (size_t)sizeof(symbol_t*)))==NULL) {
		fprintf (stderr, "malloc failed\n");
		exit(1);
	}
	zeroflag = 0;
	for (i=0; i<2; i++) {
		nrecords[i] = wam_get_vec(w, WAM_COL, t[i], &vec, 
					  XR_PTRTEMP|XR_HOSTORDER);

		if (nrecords[i]==-1) {
			copair.df[i] = 0;
			zeroflag = 1;
			fprintf (stderr, "no entry\n");
		} else {
			copair.df[i] = (nrecords[i]);
			if ((*(doc+i) = (symbol_t *)
			     calloc(nrecords[i]+1, (size_t)sizeof(symbol_t)))
			    ==NULL) {
				fprintf (stderr, "malloc 'doc+%d failed\n", i);
				exit(1);
			}
			if ((*(tf+i) = (symbol_t *)
			     calloc(nrecords[i]+1, (size_t)sizeof(symbol_t)))
			    ==NULL) {
				fprintf (stderr, "malloc 'tf+%d failed\n", i);
				exit(1);
			}
			for (r=0; r<nrecords[i]; r++) {
				xr_elem *e = &vec->elems[r];
				doc[i][r] = e->id;
				tf[i][r]  = e->freq;
			}
		}
	}
	if (zeroflag==1) {
		copair.df[2] = 0;
		copair.scr   = (double)0;
		return (0);
	}
	/*r1 = r2 = 1; /* 0Ͻ */
	r1 = r2 = 0; /* 0Ϥޤ */

	freq = 0;
	switch (calc_mode) {
	    case 10:                 /* vector calculation -cosine distance- */
		xy=xi=yi=(double)0;
		while (r1<nrecords[0] || r2<nrecords[1]) { /* ξ */
			if (r1<nrecords[0] && r2<nrecords[1]) {
				if (doc[0][r1]==doc[1][r2]) {
					xy+=(double)tf[0][r1] * (double)tf[1][r2];
					xi+=(double)tf[0][r1] * (double)tf[0][r1];
					yi+=(double)tf[1][r2] * (double)tf[1][r2];
					freq++;
					r1++;
					r2++;
				} else if (doc[0][r1]<doc[1][r2]) {
					xi+=(double)tf[0][r1] * (double)tf[0][r1];
					r1++;
				} else {
					yi+=(double)tf[1][r2] * (double)tf[1][r2];
					r2++;
				}
			} else if (r1<nrecords[0]) {
				xi+=(double)tf[0][r1] * (double)tf[0][r1];
				r1++;
			} else if (r2<nrecords[1]) {
				yi+=(double)tf[1][r2] * (double)tf[1][r2];
				r2++;
			}
		}
		if ((double)0 == (xi*yi)) {
			scr = (double)0;
		} else {
			xi = sqrt(xi);
			yi = sqrt(yi);
			scr = xy / (xi*yi);
		}
		if (threshold == (double)0) threshold = (double)0.005;
		break;

	    default:/*ɤ餫Υǡʤʤޤ*/
		while (r1<nrecords[0] && r2<nrecords[1]) { 
			/* id¤ǤȿƤ */
			if (doc[0][r1]==doc[1][r2]) {
				freq++;
				r1++;
				r2++;
			} else if (doc[0][r1]<doc[1][r2]) {
				r1++;
			} else {
				r2++;
			}
		}
		/* freq = df(t1,t2) */
		scr =  calc_co_score ((nrecords[0]), (nrecords[1]), 
				      freq, calc_mode, N);
		break;
	}
	copair.df[2] = freq;
	copair.scr   = scr;
	for (i=0; i<2; i++) {
		free(doc[i]);
		free(tf[i]);
	}
	return (1);
}

int co_scr_num=0;

/*
 * xrinfo1, xrinfo2 ơcalc_modeǶ׻ copair
 */
int calc_pair_coscr (xri1, xri2, calc_mode, N)
    XRinfo *xri1, *xri2;
    int   calc_mode, N;
{
	int i, j, r1, r2, zeroflag, freq, max1, max2, limit;
	double xy, xi, yi, bunbo;
	double scr;
	
	zeroflag=0;
	if (xri1->nrecords==-1) {
		copair.df[0]=0;
		zeroflag=1;
	}
	if (xri2->nrecords==-1) {
		copair.df[1]=0;
		zeroflag=1;
	}
	if (zeroflag==1) {
		copair.df[2]=-1;
		copair.scr = (double)0;
		return (0);
	}
	r1=r2=freq=0;
	xi = xri1->sqrsum;
	yi = xri2->sqrsum;
	/*
	max1=xri1->doc[(xri1->nrecords-2)];
	max2=xri2->doc[(xri2->nrecords-2)];
	*/
	/* ֺǸΥɥid */
	max1=xri1->doc[(xri1->nrecords-1)];
	max2=xri2->doc[(xri2->nrecords-1)];

	if (max1>max2) limit =max2;
	else           limit =max1;
	bunbo = xi*yi;
	xy=(double)0;
	while (xri1->doc[r1]<=limit && xri2->doc[r2]<=limit ) {
		while(xri1->doc[r1]<=limit && xri1->doc[r1]<xri2->doc[r2]) {
			r1++;
		}
		while(xri2->doc[r2]<=limit && xri1->doc[r1]>xri2->doc[r2]) {
			r2++;
		}
		if (xri1->doc[r1]==xri2->doc[r2]) {
			xy+=(double)xri1->tf[r1]*(double)xri2->tf[r2];
			freq++;     r1++;     r2++;
		}
	}
	switch (calc_mode) {
	    case 10:  /* vector space model? cosine distance */
		if ((double)0 == (bunbo)) {
			fprintf (stderr, "Warning: xi*yi=0\n");
			scr = (double)0;
		} else {
			scr = xy / bunbo;
		}
		if (threshold == (double)0) threshold = (double)0.005;
		break;

	    default:
	    /*
		scr = calc_co_score (xri1->nrecords-1, xri2->nrecords-1,
				     freq, calc_mode, N);
	     */
		scr = calc_co_score (xri1->nrecords, xri2->nrecords,
				     freq, calc_mode, N);
		break;
	}
	copair.df[0]= xri1->nrecords;
	copair.df[1]= xri2->nrecords;
	copair.df[2]= freq;
	copair.scr = scr;
	return (1);
}

/*  ( using )
 *  termϤ wam XRinfo˥ǡ
 */
XRinfo* get_xrinfo (term, w)
     char *term;
     WAM  *w;
{
	double tfadd;
	extern int errno;
	int j, t, zeroflag;
	XRinfo *xri;

	symbol_t id;
	ival_t elem_num;
	xr_vec *vec;

	/*
	struct wam_block *block = NULL;
	struct wam_record *rcd;
	struct xr_block *block = NULL;
	struct xr_record *rcd;
	*/

	if ((xri = (XRinfo *)calloc((size_t)1, sizeof(XRinfo)))==NULL) {
		fprintf (stderr, "malloc error for 'xri'\n");
		exit(1);
	}

	/*
	t = (int)cw_intern (term);
	xri->id  = t;
	xri->nrecords = get_wam_ptr (d, t, &block, 0);
	*/

	if (!(id=wam_name2id(w, WAM_COL, term))) {
		zeroflag=1;
		xri->id =-1;
		xri->nrecords=-1;
	} else {
		if (!(elem_num = wam_get_vec(w, WAM_COL, id, &vec, 
					     XR_PTRTEMP|XR_HOSTORDER))==-1) {
			fprintf (stderr, "ԡget_vec\n");
			exit(1);
		}
		assert (elem_num == vec->elem_num);

		xri->id = id;
		xri->nrecords=elem_num;

		tfadd=(double)0;
		if ((xri->doc = (symbol_t *)
		     calloc((xri->nrecords)+2, sizeof(symbol_t)))==NULL) {
			fprintf (stderr, "error, %d, %s\n", xri->nrecords+2,
				 strerror(errno)); exit(1);}
		if ((xri->tf  = (ival_t *)
		     calloc((xri->nrecords)+2, sizeof(ival_t)))==NULL) {
			fprintf (stderr, "error xri[].tf %d, %s\n", 
				 xri->nrecords+2, strerror(errno)); exit(1);}
		
		for (j=0; j<xri->nrecords; j++) {
			/*rcd = &block->records[j+1]; /*ǡ1Ϥޤ*/
			xr_elem *e = &vec->elems[j];
			xri->doc[j] = e->id;
			xri->tf[j]  = e->freq;
			tfadd+=(double)e->freq*(double)e->freq;
		}
		/*
		xri->doc[(xri->nrecords)-1] = (symbol_t)-1;
		*/
		xri->doc[xri->nrecords] = (symbol_t)-1;
		xri->sqrsum = sqrt(tfadd);
	}
	return (xri);
}

/*
 * CO_SCRΥ
 */
void co_scr_sort (left, right, a)
    int left, right;
    CO_SCR    a[];
{
	int     i, j;
	double  x;
	CO_SCR  s;

	if (left<right) {
		x = a[((left+right)/2)].scr;
		i = left-1; j = right+1;
		while (1) {
			while (a[++i].scr > x);
			while (a[--j].scr < x);
			if (i>=j) break;
			s    = a[i];
			a[i] = a[j];
			a[j] = s;
		}
		co_scr_sort (left, i-1,   a);
		co_scr_sort (j+1 , right, a);
	}
}

double get_max_score(calc_mode, N)
     int calc_mode, N;
{
	
	double max;

	switch (calc_mode) {
	    case 1:
		max = (double)99999999;
		break;
	    case 2:
		max = (double)N*(double)N;
		break;
	    case 3:
	    case 4:
		max = log((double)N)*(double)N;
		break;
	    case 5:
	    case 6:
		max = (double)N*(double)N;
		break;
	    case 10:
		max = (double)1.1;
		break;
	    default:
		max = log((double)N*(double)N) * (double)N;
		break;
	}
	return max;
}
